import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path

from lowball.builtins.logging.formatter import DefaultFormatter


class DefaultLoggingHandler(RotatingFileHandler):
    """
    Default log handler for Lowball applications. This handler will be added to
    the builtin flask logger. It basically is a more tightly controlled version of
    the RotatingFileHandler provided by the logging library, with the DefaultFormatter
    set as the formatter.
    """
    LOG_LEVELS = [
        logging.NOTSET,
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL
    ]

    def __init__(self, filename="lowball.log", formatter=None, log_level=logging.DEBUG, max_bytes=2**20,
                 backup_count=5):
        if not isinstance(filename, (str, Path)):
            raise TypeError("filename must be a string or a Path object")

        if formatter is None:
            formatter = {}

        if not isinstance(max_bytes, int):
            raise TypeError("max_bytes must be an int")

        if not isinstance(backup_count, int):
            raise TypeError("backup_count must be an int")

        if log_level not in self.LOG_LEVELS:
            raise ValueError(f"log_level must be one of logging library log level")

        if max_bytes < 0:
            raise ValueError("max_bytes must be zero or greater")

        if backup_count < 0:
            raise ValueError("backup_count must be zero or greater")

        super(DefaultLoggingHandler, self).__init__(filename=filename, maxBytes=max_bytes, backupCount=backup_count)
        self.setLevel(log_level)
        self.setFormatter(DefaultFormatter(**formatter))



from werkzeug.exceptions import HTTPException


class LowballException(HTTPException):
    """
    Base exception class for Lowball Exceptions
    """
    # Treat all exceptions as 500 unless explicitly overwritten
    code = 500

    # Handle Generic Exceptions
    description = "An Error Occurred. Please Check the Logs For Additional Data"

    # Since we are overwriting the base init class fulfilling the need of setting the response to None
    response = None

    def __init__(self, additional_log_data=None):
        """
        Initialize a Lowball Exception

        :param additional_log_data: additional data that can be placed in the logs but not shown to the user
        """
        # Set the optional Additional Log Data that is not shown to the user
        self.additional_log_data = additional_log_data


class InvalidTokenLifetimeException(LowballException):
    code = 400  # Bad Request
    description = "The Requested Lifetime for the Token Is Not In Range"


class NoAuthHeaderException(LowballException):
    code = 400  # Bad Request
    description = "No token provided with request"


class InvalidAuthHeaderException(LowballException):
    code = 400  # Bad Request
    description = "Authorization header improperly formatted"


class RequestNotJSONException(LowballException):
    code = 400  # Bad Request
    description = "Ill formatted request, the expected body was of type JSON"


class InvalidTokenException(LowballException):
    code = 401  # Unauthorized
    description = "Token Is Invalid"


class MalformedTokenIdException(LowballException):
    code = 400
    description = "Invalid token id format"


class ExpiredTokenException(LowballException):
    code = 401  # Unauthorized
    description = "Token Has Expired"


class InadequateRolesException(LowballException):
    code = 401  # Unauthorized
    description = "Current Token Has Inadequate Roles for Requested Action"


class AuthenticationNotInitializedException(LowballException):
    code = 503  # Service Unavailable
    description = "No Authentication Provider Present"


class InvalidCredentialsException(LowballException):
    code = 401  # Unauthorized
    description = "Invalid credentials supplied"


class MalformedAuthPackageException(LowballException):
    code = 400  # Bad Request
    description = "The Authentication Request Did Not Supply The Required Data"


class MalformedProviderPackageException(LowballException):
    code = 400  # Bad Request
    description = "The Request Did Not Supply The Required Data for the Auth Provider"


class NoAuthenticationDatabaseException(LowballException):
    code = 503
    description = "No authentication database configured for this application"


class InvalidAuthDataException(LowballException):
    description = "Unable to create token. The authentication provider returned an unrecognized response."


class InvalidRequestingUserException(LowballException):
    description = "Attempted to create a token with an invalid requesting user"


class InvalidTokenLifeException(LowballException):
    description = "Attempted to create a token where expiration is greater than configured max token life"


class ListTokensInvalidReturnTypeException(LowballException):
    description = "auth_db.list_tokens did not return a list as expected"


class ListClientsInvalidReturnTypeException(LowballException):
    description = "auth_provider.list_clients did not return a list of client data objects as expected"



class ListTokensInvalidTokenInListException(LowballException):
    description = "auth_db.list_tokens returned a list that included a non-Token object"


class TokenNotFoundException(LowballException):
    code = 404
    description = "Specified token not found"


class BadRequestException(LowballException):
    code = 400

    def __init__(self, description, additional_log_data=None):
        super(BadRequestException, self).__init__(additional_log_data)
        self.description = description


class InternalServerErrorException(LowballException):
    code = 500

    def __init__(self, description, additional_log_data=None):
        super(InternalServerErrorException, self).__init__(additional_log_data)
        self.description = description


class NotFoundException(LowballException):
    code = 404

    def __init__(self, description, additional_log_data=None):
        super(NotFoundException, self).__init__(additional_log_data)
        self.description = description


class NotImplementedException(LowballException):
    code = 501

    def __init__(self, function, additional_log_data=None):
        super(NotImplementedException, self).__init__(additional_log_data)
        self.description = f"{function} not implemented"


LOWBALL_EXCEPTIONS = [
    LowballException,
    InvalidTokenException,
    NoAuthHeaderException,
    InvalidAuthHeaderException,
    RequestNotJSONException,
    InvalidTokenException,
    ExpiredTokenException,
    InadequateRolesException,
    AuthenticationNotInitializedException,
    InvalidCredentialsException,
    MalformedAuthPackageException,
    NoAuthenticationDatabaseException,
    MalformedProviderPackageException,
    MalformedTokenIdException,
    ListClientsInvalidReturnTypeException
]


__all__ = [
    "LowballException",
    "InvalidTokenLifetimeException",
    "NoAuthHeaderException",
    "InvalidAuthHeaderException",
    "RequestNotJSONException",
    "InvalidTokenException",
    "ExpiredTokenException",
    "InadequateRolesException",
    "AuthenticationNotInitializedException",
    "InvalidCredentialsException",
    "MalformedAuthPackageException",
    "NoAuthenticationDatabaseException",
    "LOWBALL_EXCEPTIONS",
    "InvalidAuthDataException",
    "InvalidRequestingUserException",
    "InvalidTokenLifeException",
    "ListTokensInvalidTokenInListException",
    "ListTokensInvalidReturnTypeException",
    "TokenNotFoundException",
    "BadRequestException",
    "InternalServerErrorException",
    "NotFoundException",
    "NotImplementedException",
    "MalformedProviderPackageException",
    "MalformedTokenIdException",
    "ListClientsInvalidReturnTypeException"
]



import json
from unittest.mock import Mock

import pytest
from flask import Response

from lowball.builtins.response_class import LowballResponse


@pytest.fixture
def dict_return_value():
    return {"string": "value", "int": 1, "float": 1.2, "list": [1, 2, 3], "bool": False}


@pytest.fixture
def expected_dict_return(dict_return_value):
    return json.dumps(dict_return_value, separators=(',', ':'), sort_keys=True).encode() + b"\n"


@pytest.fixture(params=[
    [1, 2, 3],
    (1, 2, 3),
    {1, 2, 3},
    frozenset({1, 2, 3}),
    range(1, 4)
])
def sequence_return_value(request):
    return request.param


@pytest.fixture(params=[1, 2, 3])
def int_return_value(request):
    return request.param


@pytest.fixture(params=[1.1, 2.2, 3.3])
def float_return_value(request):
    return request.param


@pytest.fixture(params=[complex(1, 1), complex(2, 2), complex(3, 3)])
def complex_return_value(request):
    return request.param


@pytest.fixture(params=[memoryview(b"hello"), memoryview(b"world"), memoryview(b"something")])
def memoryview_return_value(request):
    return request.param


@pytest.fixture
def mocked_response_force_type(monkeypatch):
    monkeypatch.setattr(Response, "force_type", Mock(return_value=LowballResponse(response="fake_data")))


@pytest.fixture
def response_return_value():
    return Response("fake data")



import pytest
from flask import Flask

from lowball.builtins.response_class import LowballResponse


@pytest.fixture
def client_with_response_class():
    test_app = Flask("test")
    test_app.config["TESTING"] = True
    test_app.response_class = LowballResponse

    with test_app.test_request_context() as client:
        yield client



import logging
from unittest.mock import call

import flask
import pytest

from lowball import Lowball
from lowball.authentication import Authentication
from lowball.builtins import DefaultAuthProvider, DefaultAuthDB, DefaultLoggingHandler, default_error_handler, \
    LowballResponse, request_finished_log
from lowball.models.provider_models.auth_db import AuthDatabase
from lowball.models.provider_models.auth_provider import AuthProvider
from lowball.routes.builtins import status, auth


class TestLowballCoreInit:
    def test_init_raises_exception_when_incorrect_type_passed_in_for_config(self):
        with pytest.raises(TypeError):
            Lowball(config=1, logging_handler=DefaultLoggingHandler, auth_database=DefaultAuthDB,
                    auth_provider=DefaultAuthProvider, default_error_handler=default_error_handler)

    def test_passing_none_to_provider_configs_does_not_produce_errors(self, lowball_config, mocked_lowball_init,
                                                                      mocked_mkdir):
        app = Lowball(config=lowball_config, logging_handler=None, auth_provider=None, auth_database=None,
                      error_handler=None, response_class=None)

        assert app.logging_handler is None
        assert app.auth_provider is None
        assert app.auth_db is None
        assert app.error_handler == default_error_handler
        assert app.response_class == LowballResponse
        assert isinstance(app.authenticator, Authentication)

        assert len(app.logger.handlers) == 0  # This means that no handler was added during init
        flask.Flask.register_blueprint.assert_called()

        assert app.before_request_funcs[None][0].__name__ == "add_rid"

    def test_init_performs_all_necessary_actions(self, lowball_config, mocked_lowball_init, mocked_mkdir):
        app = Lowball(config=lowball_config, logging_handler=DefaultLoggingHandler, auth_database=DefaultAuthDB,
                      auth_provider=DefaultAuthProvider, error_handler=default_error_handler,
                      response_class=LowballResponse)

        assert isinstance(app.logging_handler, logging.Handler)
        assert isinstance(app.auth_provider, AuthProvider)
        assert isinstance(app.auth_db, AuthDatabase)
        assert isinstance(app.authenticator, Authentication)
        assert app.error_handler == default_error_handler
        assert app.response_class == LowballResponse

        assert app.logging_handler in app.logger.handlers
        assert app.logger.level == app.logging_handler.level
        flask.Flask.register_blueprint.assert_has_calls([call(status), call(auth)], any_order=True)

        # Assert that the `add_rid` function was added to before_request_funcs
        assert app.before_request_funcs[None][0].__name__ == "add_rid"

        flask.Flask.register_error_handler.assert_called_with(Exception, default_error_handler)
        app.register_request_finished_handler.assert_called_once_with(request_finished_log)

    @pytest.mark.parametrize("logging_handler, authentication_database, authentication_provider, response_class", [
        (1, DefaultAuthDB, DefaultAuthProvider, LowballResponse),
        (DefaultLoggingHandler, 1, DefaultAuthProvider, LowballResponse),
        (DefaultLoggingHandler, DefaultAuthDB, 1, LowballResponse),
        (DefaultLoggingHandler, DefaultAuthDB, DefaultAuthProvider, 1)
    ])
    def test_passing_incorrect_types_to_init_raises_exception(self, lowball_config, logging_handler,
                                                              authentication_database, authentication_provider,
                                                              response_class, mocked_mkdir):
        with pytest.raises(TypeError):
            Lowball(config=lowball_config, logging_handler=logging_handler, auth_provider=authentication_provider,
                    auth_database=authentication_database, response_class=response_class)


class TestLowballCoreSignalHandlers:
    # Template Rendered
    def test_template_rendered_handler_connects_single_function(self, lowball_app, signal_subscriber,
                                                                mocked_signals):
        lowball_app.register_template_rendered_handler(signal_subscriber)
        flask.template_rendered.connect.assert_called_once_with(signal_subscriber, lowball_app)

    def test_template_rendered_handler_connects_multiple_functions(self, lowball_app, signal_subscriber,
                                                                   other_signal_subscriber, mocked_signals):
        lowball_app.register_template_rendered_handler(signal_subscriber, other_signal_subscriber)
        flask.template_rendered.connect.assert_has_calls([
            call(signal_subscriber, lowball_app),
            call(other_signal_subscriber, lowball_app)
        ])

    # Before Render Template
    def test_before_render_template_handler_connects_single_function(self, lowball_app, signal_subscriber,
                                                                     mocked_signals):
        lowball_app.register_before_render_template_handler(signal_subscriber)
        flask.before_render_template.connect.assert_called_once_with(signal_subscriber, lowball_app)

    def test_before_render_template_handler_connects_multiple_functions(self, lowball_app, signal_subscriber,
                                                                        other_signal_subscriber, mocked_signals):
        lowball_app.register_before_render_template_handler(signal_subscriber, other_signal_subscriber)
        flask.before_render_template.connect.assert_has_calls([
            call(signal_subscriber, lowball_app),
            call(other_signal_subscriber, lowball_app)
        ])

    # Request Started
    def test_request_started_handler_connects_single_function(self, lowball_app, signal_subscriber,
                                                              mocked_signals):
        lowball_app.register_request_started_handler(signal_subscriber)
        flask.request_started.connect.assert_called_once_with(signal_subscriber, lowball_app)

    def test_request_started_handler_connects_multiple_functions(self, lowball_app, signal_subscriber,
                                                                 other_signal_subscriber, mocked_signals):
        lowball_app.register_request_started_handler(signal_subscriber, other_signal_subscriber)
        flask.request_started.connect.assert_has_calls([
            call(signal_subscriber, lowball_app),
            call(other_signal_subscriber, lowball_app)
        ])

    # Request Finished
    def test_request_finished_handler_connects_single_function(self, lowball_app, signal_subscriber,
                                                               mocked_signals):
        lowball_app.register_request_finished_handler(signal_subscriber)
        flask.request_finished.connect.assert_called_once_with(signal_subscriber, lowball_app)

    def test_request_finished_handler_connects_multiple_functions(self, lowball_app, signal_subscriber,
                                                                  other_signal_subscriber, mocked_signals):
        lowball_app.register_request_finished_handler(signal_subscriber, other_signal_subscriber)
        flask.request_finished.connect.assert_has_calls([
            call(signal_subscriber, lowball_app),
            call(other_signal_subscriber, lowball_app)
        ])

    # Got Request Exception
    def test_got_request_exception_handler_connects_single_function(self, lowball_app, signal_subscriber,
                                                                    mocked_signals):
        lowball_app.register_got_request_exception_handler(signal_subscriber)
        flask.got_request_exception.connect.assert_called_once_with(signal_subscriber, lowball_app)

    def test_got_request_exception_handler_connects_multiple_functions(self, lowball_app, signal_subscriber,
                                                                       other_signal_subscriber, mocked_signals):
        lowball_app.register_got_request_exception_handler(signal_subscriber, other_signal_subscriber)
        flask.got_request_exception.connect.assert_has_calls([
            call(signal_subscriber, lowball_app),
            call(other_signal_subscriber, lowball_app)
        ])

    # Request Tearing Down
    def test_request_tearing_down_handler_connects_single_function(self, lowball_app, signal_subscriber,
                                                                   mocked_signals):
        lowball_app.register_request_tearing_down_handler(signal_subscriber)
        flask.request_tearing_down.connect.assert_called_once_with(signal_subscriber, lowball_app)

    def test_request_tearing_down_handler_connects_multiple_functions(self, lowball_app, signal_subscriber,
                                                                      other_signal_subscriber, mocked_signals):
        lowball_app.register_request_tearing_down_handler(signal_subscriber, other_signal_subscriber)
        flask.request_tearing_down.connect.assert_has_calls([
            call(signal_subscriber, lowball_app),
            call(other_signal_subscriber, lowball_app)
        ])

    # Appcontext Tearing Down
    def test_appcontext_tearing_down_handler_connects_single_function(self, lowball_app, signal_subscriber,
                                                                      mocked_signals):
        lowball_app.register_appcontext_tearing_down_handler(signal_subscriber)
        flask.appcontext_tearing_down.connect.assert_called_once_with(signal_subscriber, lowball_app)

    def test_appcontext_tearing_down_handler_connects_multiple_functions(self, lowball_app, signal_subscriber,
                                                                         other_signal_subscriber, mocked_signals):
        lowball_app.register_appcontext_tearing_down_handler(signal_subscriber, other_signal_subscriber)
        flask.appcontext_tearing_down.connect.assert_has_calls([
            call(signal_subscriber, lowball_app),
            call(other_signal_subscriber, lowball_app)
        ])

    # Appcontext Pushed
    def test_appcontext_pushed_handler_connects_single_function(self, lowball_app, signal_subscriber,
                                                                mocked_signals):
        lowball_app.register_appcontext_pushed_handler(signal_subscriber)
        flask.appcontext_pushed.connect.assert_called_once_with(signal_subscriber, lowball_app)

    def test_appcontext_pushed_handler_connects_multiple_functions(self, lowball_app, signal_subscriber,
                                                                   other_signal_subscriber, mocked_signals):
        lowball_app.register_appcontext_pushed_handler(signal_subscriber, other_signal_subscriber)
        flask.appcontext_pushed.connect.assert_has_calls([
            call(signal_subscriber, lowball_app),
            call(other_signal_subscriber, lowball_app)
        ])

    # Appcontext Popped
    def test_appcontext_popped_handler_connects_single_function(self, lowball_app, signal_subscriber,
                                                                mocked_signals):
        lowball_app.register_appcontext_popped_handler(signal_subscriber)
        flask.appcontext_popped.connect.assert_called_once_with(signal_subscriber, lowball_app)

    def test_appcontext_popped_handler_connects_multiple_functions(self, lowball_app, signal_subscriber,
                                                                   other_signal_subscriber, mocked_signals):
        lowball_app.register_appcontext_popped_handler(signal_subscriber, other_signal_subscriber)
        flask.appcontext_popped.connect.assert_has_calls([
            call(signal_subscriber, lowball_app),
            call(other_signal_subscriber, lowball_app)
        ])

    # Message Flashed
    def test_message_flashed_handler_connects_single_function(self, lowball_app, signal_subscriber,
                                                              mocked_signals):
        lowball_app.register_message_flashed_handler(signal_subscriber)
        flask.message_flashed.connect.assert_called_once_with(signal_subscriber, lowball_app)

    def test_message_flashed_handler_connects_multiple_functions(self, lowball_app, signal_subscriber,
                                                                 other_signal_subscriber, mocked_signals):
        lowball_app.register_message_flashed_handler(signal_subscriber, other_signal_subscriber)
        flask.message_flashed.connect.assert_has_calls([
            call(signal_subscriber, lowball_app),
            call(other_signal_subscriber, lowball_app)
        ])

    def test_custom_signal_handler_connects_single_function(self, lowball_app, signal_subscriber, mocked_signals,
                                                            custom_signal):
        lowball_app.register_custom_signal_handler(custom_signal, signal_subscriber)
        custom_signal.connect.assert_called_once_with(signal_subscriber, lowball_app)

    def test_custom_signal_handler_connects_multiple_functions(self, lowball_app, signal_subscriber, mocked_signals,
                                                               other_signal_subscriber, custom_signal):
        lowball_app.register_custom_signal_handler(custom_signal, signal_subscriber, other_signal_subscriber)
        custom_signal.connect.assert_has_calls([
            call(signal_subscriber, lowball_app),
            call(other_signal_subscriber, lowball_app)
        ])


class TestLowballCoreAddURLRule:
    @pytest.mark.parametrize("rule, expected_result", [
        ("/", "/base/route/"),
        ("/test", "/base/route/test")
    ])
    def test_add_url_rule_properly_adds_base_route(self, lowball_app, rule, expected_result):
        def test_view_func():
            pass

        lowball_app.add_url_rule(rule, endpoint="test_view_func", view_func=test_view_func, strict_slashes=False)

        url_map = lowball_app.url_map
        rules = {route.endpoint: route.rule for route in url_map.iter_rules()}
        assert rules["test_view_func"] == expected_result



from pathlib import Path

import pytest

from lowball.models.config_models import *


@pytest.fixture(params=[
    ["string", "list", "with", "number", 1],
    ["string", "list", "with", "bool", True],
    ["string", "list", "with", "object", object()],
])
def invalid_tags(request):
    return request.param


@pytest.fixture(params=[
    "base-route-doesnt-start-with-slash"
])
def invalid_base_route(request):
    return request.param


@pytest.fixture
def auth_db_config():
    return {
        "db_host": "database.test.com",
        "db_port": 9999,
        "db_user": "root",
        "db_password": "BadPass01"
    }


@pytest.fixture
def auth_provider_config():
    return {
        "field1": 1,
        "field2": "string"
    }


@pytest.fixture
def logging_config():
    return {
        "logging_path": "/var/log/app/app.log"
    }


@pytest.fixture
def application_config():
    return {
        "field1": 1,
        "field2": "string"
    }


@pytest.fixture
def config_dict():
    return {
        "meta": {
            "name": "APP_NAME",
            "base_route": "/base/route",
            "description": "description",
            "tags": [
                "tag",
                "list"
            ]
        },
        "authentication": {
            "default_token_life": 1800,
            "max_token_life": 1800,
            "token_secret": "supersecrettokensecret"
        },
        "application": {
            "field1": 1,
            "field2": "string"
        },
        "auth_provider": {
            "field1": 1,
            "field2": "string"
        },
        "auth_db": {
            "db_host": "database.test.com",
            "db_port": 9999,
            "db_user": "root",
            "db_password": "BadPass01"
        },
        "logging": {
            "logging_path": "/var/log/app/app.log"
        }
    }


@pytest.fixture(params=[1, "string", [1, 2, 3], ("tuple",), {1, 2, 3}])
def not_dict(request):
    return request.param


@pytest.fixture(params=[
    {"meta": {"name": "APP_NAME", "base_route": "/base/route", "description": "description", "tags": ["tag", "list"]},
     "authentication": {"default_token_life": 1800, "max_token_life": 1800, "token_secret": "supersecrettokensecret", "ignore_auth": False},
     "application": {"field1": 1, "field2": "string"}, "auth_provider": {"field1": 1, "field2": "string"},
     "auth_db": {"db_host": "database.test.com", "db_port": 9999, "db_user": "root", "db_password": "BadPass01"},
     "logging": {"logging_path": "/var/log/app/app.log"}},
    {"meta": {"name": "APP_NAME", "base_route": "/base/route", "description": "description", "tags": ["tag", "list"]},
     "authentication": {"default_token_life": 1800, "max_token_life": 1800, "token_secret": "supersecrettokensecret", "ignore_auth": False},
     "application": {}, "auth_provider": {}, "auth_db": {}, "logging": {}}
])
def json_config_object(request):
    return request.param


@pytest.fixture(params=[
    "/path/to/config.json",
    Path("/path/to/config.json")
])
def json_config_path(request):
    return request.param


@pytest.fixture
def json_file_read_data():
    return '{\n  "meta": {\n    "name": "APP_NAME",\n    "base_route": "/base/route",\n    "description": "description",\n    "tags": [\n      "tag",\n      "list"\n    ]\n  },\n  "application": {\n    "field1": 1,\n    "field2": "string"\n  },\n  "auth_provider": {\n    "field1": 1,\n    "field2": "string"\n  },\n  "auth_db": {\n    "db_host": "database.test.com",\n    "db_port": 9999,\n    "db_user": "root",\n    "db_password": "BadPass01"\n  },\n  "logging": {\n    "logging_path": "/var/log/app/app.log"\n  },\n  "authentication": {\n    "default_token_life": 1800,\n    "max_token_life": 1800,\n    "token_secret": "supersecrettokensecret",\n    "ignore_auth": false\n  }\n}'


@pytest.fixture(params=[
    "/path/to/config.yaml",
    Path("/path/to/config.yaml")
])
def yaml_config_path(request):
    return request.param


@pytest.fixture
def yaml_file_read_data():
    return 'meta:\n  name: APP_NAME\n  base_route: /base/route\n  description: description\n  tags:\n  - tag\n  - list\napplication:\n  field1: 1\n  field2: string\nauth_provider:\n  field1: 1\n  field2: string\nauth_db:\n  db_host: database.test.com\n  db_port: 9999\n  db_user: root\n  db_password: BadPass01\nlogging:\n  logging_path: /var/log/app/app.log\nauthentication:\n  default_token_life: 1800\n  max_token_life: 1800\n  token_secret: supersecrettokensecret\n  ignore_auth: false\n'


@pytest.fixture(params=["1", "1234", "1234567890"])
def int_as_string(request):
    return request.param


@pytest.fixture(params=[{}, [], object(), True, 1.1])
def not_like_int(request):
    return request.param


@pytest.fixture(params=[{}, [], object(), 0, 1.1, "string"])
def not_bool(request):
    return request.param


@pytest.fixture
def auth_config_dict(default_token_life, max_token_life, token_secret):
    return {
        "default_token_life": int(default_token_life),
        "max_token_life": int(max_token_life),
        "token_secret": token_secret,
    }



from datetime import datetime
from lowball.models.authentication_models import ClientData
import pytest
_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


@pytest.fixture(params=[
    ["string", "list", "with", "number", 1],
    ["string", "list", "with", "bool", True],
    ["string", "list", "with", "object", object()],
])
def invalid_roles(request):
    return request.param


@pytest.fixture
def roles():
    return ["lead", "analyst"]


@pytest.fixture(params=[datetime(2020, 1, 1, 0, 0, 0, 0), "2020-01-01 00:00:00"])
def created(request):
    return request.param


@pytest.fixture(params=[datetime(2020, 2, 1, 0, 0, 0, 0), "2020-02-01 00:00:00"])
def expiration(request):
    return request.param


@pytest.fixture
def issued_by():
    return "issuer"


@pytest.fixture
def auth_data_dict(username, roles):
    return {
        "username": username,
        "roles": roles
    }


@pytest.fixture
def token_dict(username, roles, created, expiration, issued_by, token_id):
    return {
        "cid": username,
        "r": roles,
        "cts": datetime.strftime(created, _DATE_FORMAT) if isinstance(created, datetime) else str(created),
        "ets": datetime.strftime(expiration, _DATE_FORMAT) if isinstance(expiration, datetime) else str(expiration),
        "rcid": issued_by,
        "tid": token_id
    }


@pytest.fixture(params=[1, True, object(), [1, 2, 3]])
def not_string(request):
    return request.param


@pytest.fixture(params=["bad_date_string", "1606942066224648", "20200101000000"])
def bad_date_string(request):
    return request.param


@pytest.fixture(params=[1, 1.1, True, object(), {}, []])
def invalid_username(request):
    return request.param


@pytest.fixture
def valid_username():
    return "username"


@pytest.fixture
def user(valid_username, roles):
    return ClientData(client_id=valid_username, roles=roles)


@pytest.fixture
def expected_user_to_dict_return(valid_username, roles):
    return {
        "client_id": valid_username,
        "roles": roles
    }



import logging
import re
import uuid

from flask import Flask, request, Response
from flask.signals import *

from lowball.authentication import Authentication
from lowball.builtins import DefaultLoggingHandler, DefaultAuthDB, DefaultAuthProvider, default_error_handler, \
    LowballResponse, request_finished_log
from lowball.models.config_models import Config
from lowball.models.provider_models import AuthDatabase, AuthProvider
from lowball.routes.builtins import auth, status


class Lowball(Flask):
    """Core class of a Lowball application.

    This class is an extension of Flask. In addition to the services that Flask provides,
    Lowball controls and configures:
        - Loading YAML configurations and applying them to the application
        - Configuring (if any) authentication providers
        - Sets up logging for the service
        - Sets the default error handler for the application

    :type lowball_config: Config
    :param lowball_config: Config object for the application
    :type logging_handler: type
    :param logging_handler: The handler object for logging in the application -> This will be instantiated on init
    :type auth_database: type
    :param auth_database: The authentication database for the application -> This will be instantiated on init
    :type auth_provider: type
    :param auth_provider: The authentication provider for the application -> This will be instantiated on init
    :type authenticator: Authentication
    :param authenticator: attribute responsible for token operations
    :type error_handler: function
    :param error_handler: The error handler function that will be set as the default
                          error handler for the app
    """
    def __init__(self, config=Config(), logging_handler=DefaultLoggingHandler, auth_database=DefaultAuthDB,
                 auth_provider=DefaultAuthProvider, error_handler=default_error_handler, response_class=LowballResponse,
                 **kwargs):
        if not isinstance(config, Config):
            raise TypeError("config must be a Config object")

        self.lowball_config = config

        super(Lowball, self).__init__(import_name=self.lowball_config.meta.name, **kwargs)

        if logging_handler is None:
            self.logging_handler = None
        elif not issubclass(logging_handler, logging.Handler):
            raise TypeError("logging_handler must subclass logging.Handler")
        else:
            self.logging_handler = logging_handler(**self.lowball_config.logging)
            self.logger.setLevel(self.logging_handler.level)
            self.logger.addHandler(self.logging_handler)

        if auth_provider is None:
            self.auth_provider = None
        elif not issubclass(auth_provider, AuthProvider):
            raise TypeError("auth provide muse be a subclass of AuthProvider")
        else:
            self.auth_provider = auth_provider(**self.lowball_config.auth_provider)

        if auth_database is None:
            self.auth_db = None
        elif not issubclass(auth_database, AuthDatabase):
            raise TypeError("auth_database must be a subclass of AuthDatabase")
        else:
            self.auth_db = auth_database(**self.lowball_config.auth_db)

        if response_class is None:
            self.response_class = LowballResponse
        elif not issubclass(response_class, Response):
            raise TypeError("response_class must be a subclass of flask.Response")
        else:
            self.response_class = response_class

        if error_handler is None:
            self.error_handler = default_error_handler
        else:
            self.error_handler = error_handler

        self.authenticator = Authentication(config=self.lowball_config.authentication)

        self.setup_before_request()
        self.register_blueprint(status)
        self.register_blueprint(auth)

        self.register_request_finished_handler(request_finished_log)

        self.register_error_handler(Exception, self.error_handler)

    def register_template_rendered_handler(self, *subscribers):
        """Connect 1 to n subscribers to the template_rendered signal."""
        for subscriber in subscribers:
            template_rendered.connect(subscriber, self)

    def register_before_render_template_handler(self, *subscribers):
        """Connect 1 to n subscribers to the before_render_template signal."""
        for subscriber in subscribers:
            before_render_template.connect(subscriber, self)

    def register_request_started_handler(self, *subscribers):
        """Connect 1 to n subscribers to the request_started signal."""
        for subscriber in subscribers:
            request_started.connect(subscriber, self)

    def register_request_finished_handler(self, *subscribers):
        """Connect 1 to n subscribers to the request_finished signal."""
        for subscriber in subscribers:
            request_finished.connect(subscriber, self)

    def register_got_request_exception_handler(self, *subscribers):
        """Connect 1 to n subscribers to the got_request_exception signal."""
        for subscriber in subscribers:
            got_request_exception.connect(subscriber, self)

    def register_request_tearing_down_handler(self, *subscribers):
        """Connect 1 to n subscribers to the request_tearing_down signal."""
        for subscriber in subscribers:
            request_tearing_down.connect(subscriber, self)

    def register_appcontext_tearing_down_handler(self, *subscribers):
        """Connect 1 to n subscribers to the appcontext_tearing_down signal."""
        for subscriber in subscribers:
            appcontext_tearing_down.connect(subscriber, self)

    def register_appcontext_pushed_handler(self, *subscribers):
        """Connect 1 to n subscribers to the appcontext_pushed signal."""
        for subscriber in subscribers:
            appcontext_pushed.connect(subscriber, self)

    def register_appcontext_popped_handler(self, *subscribers):
        """Connect 1 to n subscribers to the appcontext_popped signal."""
        for subscriber in subscribers:
            appcontext_popped.connect(subscriber, self)

    def register_message_flashed_handler(self, *subscribers):
        """Connect 1 to n subscribers to the message_flashed signal."""
        for subscriber in subscribers:
            message_flashed.connect(subscriber, self)

    def register_custom_signal_handler(self, signal, *subscribers):
        """Connect 1 to n subscribers to a custom signal signal."""
        for subscriber in subscribers:
            signal.connect(subscriber, self)

    def add_url_rule(self, rule, endpoint=None, view_func=None, provide_automatic_options=None, **options):
        """Ensure that base route is added to all urls.

        This is an override of the flask add_url_rule which allows setting of the "base" route as specified in the
        configuration.
        """
        route = f"/{self.lowball_config.meta.base_route}/{rule}"
        route = re.sub("/{2,}", "/", route)  # squeeze any number of slashes into one slash

        super(Lowball, self).add_url_rule(route, endpoint, view_func, provide_automatic_options, **options)

    def setup_before_request(self):
        @self.before_request
        def add_rid():
            """Add a UUID to each request."""
            request.rid = str(uuid.uuid4())




import pytest
from unittest.mock import mock_open, patch

from lowball.models.config_models import *


class TestMetaConfig:
    @pytest.mark.parametrize("name, base_route, description, tags", [
        (1, "/base/route", "description", ["list", "of", "tags"]),
        ("APP_NAME", 1, "description", ["list", "of", "tags"]),
        ("APP_NAME", "/base/route", 1, ["list", "of", "tags"]),
        ("APP_NAME", "/base/route", "description", 1)
    ])
    def test_init_raises_type_error_with_incorrect_types_passed_in(self, name, base_route, description, tags):
        with pytest.raises(TypeError):
            MetaConfig(name=name, base_route=base_route, description=description, tags=tags)

    def test_init_raises_value_error_with_invalid_tag_values(self, invalid_tags):
        with pytest.raises(ValueError):
            MetaConfig(name="APP_NAME", base_route="/base/route", description="description", tags=invalid_tags)

    def test_init_raised_value_error_with_invalid_base_route_values(self, invalid_base_route):
        with pytest.raises(ValueError):
            MetaConfig(name="APP_NAME", base_route=invalid_base_route, description="description", tags=["tag", "list"])

    def test_to_dict_method_formats_data_properly(self):
        test_meta = MetaConfig(name="APP_NAME", base_route="/base/route", description="description", tags=["tag", "list"])
        assert test_meta.to_dict() == {"name": "APP_NAME", "base_route": "/base/route", "description": "description", "tags": ["tag", "list"]}

    def test_tags_can_be_set_to_none_on_init(self):
        test_meta = MetaConfig(name="APP_NAME", base_route="/base/route", description="description", tags=None)
        assert test_meta.to_dict() == {"name": "APP_NAME", "base_route": "/base/route", "description": "description",
                                       "tags": []}


class TestAuthenticationConfig:
    @pytest.mark.parametrize("default_token_life, max_token_life, secret", [
        ("not an int", 31536000, "supersecrettokensecret"),
        (1800, "not an int", "supersecrettokensecret"),
        (1800, 31536000, 1),
    ])
    def test_init_raises_type_error_with_incorrect_types_passed_in(self, default_token_life, max_token_life, secret):
        with pytest.raises(TypeError):
            AuthenticationConfig(default_token_life=default_token_life, max_token_life=max_token_life,
                                 token_secret=secret)

    def test_setting_default_token_life_to_bad_value_after_init_raises_type_error(self, default_token_life,
                                                                                  max_token_life, token_secret,
                                                                                  not_like_int):
        test_auth = AuthenticationConfig(default_token_life=default_token_life, max_token_life=max_token_life,
                                           token_secret=token_secret)
        with pytest.raises(TypeError):
            test_auth.default_token_life = not_like_int

    def test_setting_max_token_life_to_bad_value_after_init_raises_type_error(self, default_token_life, max_token_life,
                                                                              token_secret, not_like_int):
        test_auth = AuthenticationConfig(default_token_life=default_token_life, max_token_life=max_token_life,
                                           token_secret=token_secret)
        with pytest.raises(TypeError):
            test_auth.max_token_life = not_like_int

    def test_setting_token_secret_to_bad_value_after_init_raises_type_error(self, default_token_life, max_token_life,
                                                                            token_secret, not_string):
        test_auth = AuthenticationConfig(default_token_life=default_token_life, max_token_life=max_token_life,
                                           token_secret=token_secret)
        with pytest.raises(TypeError):
            test_auth.token_secret = not_string

    def test_to_dict_method_formats_data_properly(self, default_token_life, max_token_life, token_secret,
                                                  auth_config_dict):
        test_auth = AuthenticationConfig(default_token_life=default_token_life, max_token_life=max_token_life,
                                           token_secret=token_secret)
        assert test_auth.to_dict() == auth_config_dict


class TestConfig:
    @pytest.mark.parametrize("meta, authentication, application, auth_provider, auth_db, logging", [
        (1, AuthenticationConfig(default_token_life=1800, max_token_life=3600, token_secret="supersecret", ignore_auth=False), {}, {}, {}, {}),
        (MetaConfig(name="APP_NAME", base_route="/base/route", description="description", tags=["tag", "list"]), 1, {}, {}, {}, {}),
        (MetaConfig(name="APP_NAME", base_route="/base/route", description="description", tags=["tag", "list"]), AuthenticationConfig(default_token_life=1800, max_token_life=3600, token_secret="supersecret"), 1, {}, {}, {}),
        (MetaConfig(name="APP_NAME", base_route="/base/route", description="description", tags=["tag", "list"]), AuthenticationConfig(default_token_life=1800, max_token_life=3600, token_secret="supersecret"), {}, 1, {}, {}),
        (MetaConfig(name="APP_NAME", base_route="/base/route", description="description", tags=["tag", "list"]), AuthenticationConfig(default_token_life=1800, max_token_life=3600, token_secret="supersecret"), {}, {}, 1, {}),
        (MetaConfig(name="APP_NAME", base_route="/base/route", description="description", tags=["tag", "list"]), AuthenticationConfig(default_token_life=1800, max_token_life=3600, token_secret="supersecret"), {}, {}, {}, 1)
    ])
    def test_init_raises_type_error_with_incorrect_types_passed_in(self, meta, authentication, application,
                                                                   auth_provider, auth_db, logging):
        with pytest.raises(TypeError):
            Config(meta=meta, authentication=authentication, application=application, auth_provider=auth_provider,
                   auth_db=auth_db, logging=logging)

    def test_to_dict_returns_proper_data(self, auth_db_config, auth_provider_config, logging_config, application_config,
                                         meta_config, authentication_config, config_dict):
        test_config = Config(meta=meta_config, authentication=authentication_config, application=application_config,
                               auth_provider=auth_provider_config, auth_db=auth_db_config, logging=logging_config)
        assert test_config.to_dict() == config_dict

    def test_attribute_access_returns_proper_data_type(self, auth_db_config, auth_provider_config, logging_config,
                                                       application_config, meta_config, authentication_config):
        test_config = Config(meta=meta_config, authentication=authentication_config, application=application_config,
                               auth_provider=auth_provider_config, auth_db=auth_db_config, logging=logging_config)

        assert isinstance(test_config.meta, MetaConfig)
        assert isinstance(test_config.authentication, AuthenticationConfig)


class TestLoadingConfig:
    def test_loading_config_from_object_raises_type_error_when_not_dict_passed_in(self, not_dict):
        with pytest.raises(TypeError):
            config_from_object(not_dict)

    def test_loading_config_from_object_returns_proper_config_object(self, json_config_object):
        test_config = config_from_object(json_config_object)
        assert isinstance(test_config, Config)
        assert test_config.meta.to_dict() == json_config_object["meta"]
        assert test_config.application == json_config_object["application"]
        assert test_config.auth_provider == json_config_object["auth_provider"]
        assert test_config.auth_db == json_config_object["auth_db"]
        assert test_config.logging == json_config_object["logging"]

    def test_loading_config_from_file_file_raises_type_error_when_pathlike_object_not_passed_in(self, not_pathlike):
        with pytest.raises(TypeError):
            config_from_file(not_pathlike)

    def test_loading_config_from_json_file_returns_proper_config_object(self, json_config_path, json_file_read_data, config_dict):
        mocked_open_func = mock_open(read_data=json_file_read_data)

        with patch("lowball.models.config_models.config.open", mocked_open_func):
            test_config = config_from_file(json_config_path)

        mocked_open_func.assert_any_call(json_config_path, "r")
        assert isinstance(test_config, Config)
        assert config_dict == test_config.to_dict()


    def test_loading_config_from_yaml_file_returns_proper_config_object(self, yaml_config_path, yaml_file_read_data, config_dict):
        mocked_open_func = mock_open(read_data=yaml_file_read_data)

        with patch("lowball.models.config_models.config.open", mocked_open_func):
            test_config = config_from_file(yaml_config_path)

        mocked_open_func.assert_any_call(yaml_config_path, "r")
        assert isinstance(test_config, Config)
        assert config_dict == test_config.to_dict()



import logging
import pytest

from logging.handlers import RotatingFileHandler

from lowball.builtins.logging.logging_handler import DefaultLoggingHandler


class TestDefaultLoggingHandler:
    @pytest.mark.parametrize("filename, log_level, max_bytes, backup_count, formatter", [
        (1, logging.DEBUG, 2**20, 5, {"date_format": "%Y-%m-%d %H:%M:%S.%fUTC"}),
        ("/path/to/file", logging.DEBUG, "not an int", 5, {"date_format": "%Y-%m-%d %H:%M:%S.%fUTC"}),
        ("/path/to/file", logging.DEBUG, 2**20, "not an int", {"date_format": "%Y-%m-%d %H:%M:%S.%fUTC"}),
        ("/path/to/file", logging.DEBUG, 2**20, 5, "Not a dict")
    ])
    def test_passing_bad_values_at_init_raises_exception(self, filename, log_level, max_bytes, backup_count, formatter,
                                                         mock_open_file):
        with pytest.raises(TypeError):
            DefaultLoggingHandler(filename=filename, log_level=log_level, max_bytes=max_bytes,
                                  backup_count=backup_count, formatter=formatter)

    def test_init_sets_formatter_and_log_level(self, filename, log_level, max_bytes, backup_count, formatter,
                                               monkeypatched_setters, mock_open_file):
        DefaultLoggingHandler(filename=filename, log_level=log_level, max_bytes=max_bytes, backup_count=backup_count,
                              formatter=formatter)
        RotatingFileHandler.setFormatter.assert_called_once()
        RotatingFileHandler.setLevel.assert_called_once_with(log_level)

    def test_passing_bad_log_level_raises_exception(self, filename, bad_log_level, max_bytes, backup_count, formatter,
                                                    mock_open_file):
        with pytest.raises(ValueError):
            DefaultLoggingHandler(filename=filename, log_level=bad_log_level, max_bytes=max_bytes,
                                  backup_count=backup_count, formatter=formatter)

    def test_passing_bad_max_bytes_raises_exception(self, filename, log_level, bad_max_bytes, backup_count, formatter,
                                                    mock_open_file):
        with pytest.raises(ValueError):
            DefaultLoggingHandler(filename=filename, log_level=log_level, max_bytes=bad_max_bytes,
                                  backup_count=backup_count, formatter=formatter)

    def test_passing_bad_backup_count_raises_exception(self, filename, log_level, max_bytes, bad_backup_count,
                                                       formatter,mock_open_file):
        with pytest.raises(ValueError):
            DefaultLoggingHandler(filename=filename, log_level=log_level, max_bytes=max_bytes,
                                  backup_count=bad_backup_count, formatter=formatter)



import builtins
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from unittest.mock import mock_open, Mock

import pytest
from ruamel.yaml import YAML

from lowball.builtins.providers import DefaultAuthDB, DefaultAuthProvider, DefaultAuthPackage
from lowball.models.authentication_models import Token, ClientData
from lowball.models.provider_models.auth_provider import AuthPackage
import json

@pytest.fixture(params=[
    "/opt/lowball/tokens",
    Path("/opt/lowball/tokens")
])
def token_path(request):
    return request.param


@pytest.fixture
def token_file_contents(token):
    return json.dumps(token.to_dict())


@pytest.fixture
def token_file(monkeypatch, token_file_contents):
    mock_file = mock_open(read_data=token_file_contents)
    monkeypatch.setattr(builtins, "open", mock_file)


@pytest.fixture
def corrupted_token_file(monkeypatch, token_file_contents):
    mock_file = mock_open(read_data="Corrupted Data")
    monkeypatch.setattr(builtins, "open", mock_file)


@pytest.fixture
def mock_open_file(monkeypatch, token_file_contents):
    mock_file = mock_open()
    monkeypatch.setattr(builtins, "open", mock_file)
    return mock_file


@pytest.fixture
def mocked_yaml_dump(monkeypatch):
    monkeypatch.setattr(YAML, "dump", Mock())


@pytest.fixture
def mock_nonexistent_path(monkeypatch):
    monkeypatch.setattr(Path, "exists", Mock(return_value=False))


@pytest.fixture
def mock_existent_path(monkeypatch):
    monkeypatch.setattr(Path, "exists", Mock(return_value=True))


@pytest.fixture
def mock_path_unlink(monkeypatch):
    monkeypatch.setattr(Path, "unlink", Mock())


@pytest.fixture
def mock_unlink_non_existent_file(monkeypatch):
    mock = Mock()
    mock.side_effect = FileNotFoundError
    monkeypatch.setattr(Path, "unlink", mock)


@pytest.fixture
def default_auth_db(token_path, mocked_mkdir):
    return DefaultAuthDB(token_path=token_path)


@pytest.fixture
def mock_default_auth_db_internals(monkeypatch, token):
    monkeypatch.setattr(DefaultAuthDB, "_dump_token", Mock(return_value=True))
    monkeypatch.setattr(DefaultAuthDB, "_load_token", Mock(return_value=token))
    monkeypatch.setattr(DefaultAuthDB, "_delete_token", Mock(return_value=True))


@pytest.fixture(params=["token", "token_id"])
def token_identifier(request, token_id, token):
    return {"token": token, "token_id": token_id}[request.param]


@pytest.fixture
def mock_internal_load_non_existent_token(monkeypatch):
    monkeypatch.setattr(DefaultAuthDB, "_load_token", Mock(return_value=None))


@pytest.fixture
def mock_internal_revoke_token(monkeypatch):
    monkeypatch.setattr(DefaultAuthDB, "_delete_token", Mock())


@pytest.fixture
def mock_iterdir(monkeypatch, default_auth_db, token_id1, token_id2, token_id3, token_id4, token_id5):
    file_list = [
        default_auth_db.token_path.joinpath(token_id1),
        default_auth_db.token_path.joinpath(token_id2),
        default_auth_db.token_path.joinpath(token_id3),
        default_auth_db.token_path.joinpath(token_id4),
        default_auth_db.token_path.joinpath(token_id5),
    ]
    monkeypatch.setattr(Path, "iterdir", Mock(return_value=iter(file_list)))


@pytest.fixture
def mock_load_multiple_tokens_internal(monkeypatch, token, default_auth_db, token_id3, token_id4):
    def send_back_token(token_id):
        if token_id in [default_auth_db.token_path.joinpath(token_id3),
                        default_auth_db.token_path.joinpath(token_id4)]:
            return None
        else:
            token.token_id = token_id.name
            return token

    monkeypatch.setattr(DefaultAuthDB, "_load_token", Mock(wraps=send_back_token))


@pytest.fixture
def expected_list_tokens(token, token_id1, token_id2, token_id5):
    token_list = []

    for token_id in [token_id1, token_id2,
                     token_id5]:
        token.token_id = token_id
        token_list.append(token)

    return token_list


@pytest.fixture
def mock_list_tokens(monkeypatch, token_id1, token_id2, token_id3, token_id5, token_id4):
    mock_return = Mock()
    mock_return.return_value = [
        Token(cid="username", r=["admin", "lead", "analyst"], cts=datetime(2020, 1, 1),
              ets=datetime(2020, 2, 1), rcid="username", tid=token_id1),
        Token(cid="other_user", r=["admin"], cts=datetime(2020, 1, 1), ets=datetime(2020, 2, 1),
              rcid="username", tid=token_id2),
        Token(cid="username", r=["admin", "lead"], cts=datetime(2019, 1, 1),
              ets=datetime(2019, 2, 1), rcid="username", tid=token_id5),
        Token(cid="other_user", r=[], cts=datetime(2020, 1, 1), ets=datetime(2020, 2, 1),
              rcid="username", tid=token_id3),
        Token(cid="username", r=["analyst"], cts=datetime(2019, 1, 1), ets=datetime(2019, 2, 1),
              rcid="username", tid=token_id4)
    ]

    monkeypatch.setattr(DefaultAuthDB, "list_tokens", mock_return)


@pytest.fixture
def expected_filter_by_user_result(token_id1, token_id5, token_id4):
    return [
        Token(cid="username", r=["admin", "lead", "analyst"], cts=datetime(2020, 1, 1),
              ets=datetime(2020, 2, 1), rcid="username", tid=token_id1),
        Token(cid="username", r=["admin", "lead"], cts=datetime(2019, 1, 1),
              ets=datetime(2019, 2, 1), rcid="username", tid=token_id5),
        Token(cid="username", r=["analyst"], cts=datetime(2019, 1, 1), ets=datetime(2019, 2, 1),
              rcid="username", tid=token_id4)
    ]


@pytest.fixture
def expected_filter_by_role_result(token_id1, token_id4):
    return [
        Token(cid="username", r=["admin", "lead", "analyst"], cts=datetime(2020, 1, 1),
              ets=datetime(2020, 2, 1), rcid="username", tid=token_id1),
        Token(cid="username", r=["analyst"], cts=datetime(2019, 1, 1), ets=datetime(2019, 2, 1),
              rcid="username", tid=token_id4)
    ]


@pytest.fixture
def default_auth_provider():
    return DefaultAuthProvider(username="admin", password="nimda")


class MalformedAuthPackage(AuthPackage):
    def __init__(self):
        super(MalformedAuthPackage, self).__init__()


@pytest.fixture(params=[1, True, [], {"username": "admin", "password": "nimda"}, object()])
def malformed_auth_package(request):
    return request.param


@pytest.fixture(params=[
    {"username": "notadmin", "password": "nimda"},
    {"username": "admin", "password": "notnimda"},
])
def invalid_auth_package(request):
    return DefaultAuthPackage(**request.param)


@pytest.fixture
def default_auth_package():
    return DefaultAuthPackage(username="admin", password="nimda")


@pytest.fixture
def mkdir_raises_permission_error(monkeypatch):
    monkeypatch.setattr(Path, "mkdir", Mock(side_effect=PermissionError))


@pytest.fixture(params=[
    "/token/id/forward/slash",
    "\\token\\path\\back\\slash",
    "4bdad277-16b5-4d9c-a081\\5856d4f348b6",
    "4bdad277-16b5-4d9c-a081/5856d4f348b6"
])
def pathlike_token_id(request):
    return request.param


@pytest.fixture
def pathlike_token(token, pathlike_token_id):
    new_token = deepcopy(token)
    new_token._token_id = pathlike_token_id
    return new_token


@pytest.fixture(params=[
    "admin",
    "user",
    "weirdo",
    "blob",
    1,
    None,
    "flop"
])
def get_client_expected_outcomes(request, admin_role):
    client_id = request.param

    if client_id == "admin":
        return client_id, ClientData(client_id="admin", roles=admin_role)
    else:
        return client_id, None




from flask import jsonify, g
from werkzeug.exceptions import HTTPException

from lowball.exceptions import LowballException


def default_error_handler(exception):
    """Default Error Handler for Lowball

    This handler takes the exception information and converts it to json and sets the g repsonse_is_exception
    to true, which will affect log handling down the chain.
    """
    # Build the initial response with the body containing the error message
    response = jsonify(message=str(exception))

    # If its a recognized type of exception handle it accordingly if not 500
    if isinstance(exception, HTTPException):
        response.status_code = exception.code

        if isinstance(exception, LowballException):
            g.response_is_exception = True

            g.response_exception_log_data = exception.additional_log_data
        elif exception.code > 499:
            g.response_is_exception = True

            g.response_exception_log_data = {"error_type": str(type(exception)), "error_msg": str(exception)}

    else:
        g.response_is_exception = True
        response.status_code = 500
        g.response_exception_log_data = {"error_type": str(type(exception)), "error_msg": str(exception)}

    return response



import pytest
from werkzeug.exceptions import HTTPException

from lowball.exceptions import LOWBALL_EXCEPTIONS


@pytest.fixture(params=[
    Exception("test exception"),
    TypeError("test exception"),
    ValueError("test exception"),
    AttributeError("test exception")
])
def basic_exception(request):
    return request.param


@pytest.fixture(params=[exception("test exception") for exception in LOWBALL_EXCEPTIONS])
def lowball_exception(request):
    return request.param


http_exception1 = HTTPException("test exception")
http_exception1.code = 400
http_exception2 = HTTPException("test exception")
http_exception2.code = 401
http_exception3 = HTTPException("test exception")
http_exception3.code = 404


@pytest.fixture(params=[
    http_exception1,
    http_exception2,
    http_exception3
])
def http_exception_lt_500(request):
    return request.param


http_exception4 = HTTPException("test exception")
http_exception4.code = 500
http_exception5 = HTTPException("test exception")
http_exception5.code = 501
http_exception6 = HTTPException("test exception")
http_exception6.code = 503


@pytest.fixture(params=[
    http_exception4,
    http_exception5,
    http_exception6
])
def http_exception_gte_500(request):
    return request.param



import json
import logging

from copy import deepcopy
from datetime import datetime
from flask import g, request


class DefaultFormatter(logging.Formatter):
    """
    Default formatter for Lowball applications. This formatter is designed to log in a
    JSON format, and will log different data depending on the level of log that is being
    logged. Debug logs have more data than Info logs, for example. It also handles some
    renaming of the fields, and stores data that the user adds to the log in a field
    called `additional`.

    :type date_format: str
    :param date_format: the strftime format string used to format datetime strings for
                        log timestamps
    """
    DEFAULT_LOG_FIELDS = [
        "msg",
        "name",
        "args",
        "timestamp",
        "level",
        "additional"
    ]

    VERBOSE_LOG_FIELDS = DEFAULT_LOG_FIELDS + [
        "filename",
        "func_name",
        "line_number",
        "module",
        "pathname",
        "process",
        "process_name",
        "thread_name",
        "thread",
    ]

    EXCEPTION_ERROR_FIELDS = VERBOSE_LOG_FIELDS + [
        "exc_info",
        "stack_info"
    ]

    LOG_RECORD_FIELDS = [
        'args', 'asctime', 'created', 'exc_info', 'exc_text', 'filename', 'funcName', 'levelname', 'levelno', 'lineno',
        'module', 'msecs', 'message', 'msg', 'name', 'pathname', 'process', 'processName', 'relativeCreated',
        'stack_info', 'thread', 'threadName'
    ]

    RECORD_ATTRIBUTE_TRANSFORM = [
        ("funcName", "func_name"),
        ("lineno", "line_number"),
        ("levelname", "level"),
        ("processName", "process_name"),
        ("threadName", "thread_name")
    ]

    DEFAULT_LOG_LEVELS = ["NOTSET", "INFO", "WARNING"]

    EXCEPTION_ERROR_LOG_LEVELS = ["ERROR", "CRITICAL"]

    def __init__(self, date_format="%Y-%m-%d %H:%M:%S.%fUTC", **kwargs):
        self.date_format = date_format
        super(DefaultFormatter, self).__init__(datefmt=self.date_format)

    @property
    def date_format(self):
        return self._date_format

    @date_format.setter
    def date_format(self, value):
        if not isinstance(value, str):
            raise TypeError("dated_format must be a string")
        self._date_format = value

    def _format_time(self, timestamp):
        """Format a timestamp into the proper date string

        :type timestamp: float
        :param timestamp: UTC epoch timestamp
        :rtype: str
        :return: date time string
        """
        dt = datetime.utcfromtimestamp(timestamp)
        return dt.strftime(self.date_format)

    def _fix_attribute_names(self, record: logging.LogRecord):
        """Transform record field names from camel case to underscored."""
        for attribute, transform in self.RECORD_ATTRIBUTE_TRANSFORM:
            value = record.__getattribute__(attribute)
            record.__setattr__(transform, value)

        return record

    def _get_additional(self, record: logging.LogRecord):
        """Map all record fields that are not default record fields into their own dictionary."""
        additional = {k: v for k, v in record.__dict__.items() if k not in self.LOG_RECORD_FIELDS}
        return additional

    def format(self, record: logging.LogRecord):
        """Format a log record.

        This method takes the log record and transforms it into a properly formatted
        JSON string.

        :type record: logging.LogRecord
        :param record: Log record generated by logger
        :rtype: str
        :return: Log in JSON format
        """
        copied = deepcopy(record)
        copied.additional = self._get_additional(copied)
        copied.timestamp = self._format_time(copied.created)

        copied = self._fix_attribute_names(copied)

        if copied.level in self.DEFAULT_LOG_LEVELS:
            keep_fields = self.DEFAULT_LOG_FIELDS
        elif copied.level in self.EXCEPTION_ERROR_LOG_LEVELS:
            keep_fields = self.EXCEPTION_ERROR_FIELDS
            if copied.exc_info:
                copied.exc_info = self.formatException(copied.exc_info)
            if copied.stack_info:
                copied.stack_info = self.formatStack(copied.stack_info)
        else:
            keep_fields = self.VERBOSE_LOG_FIELDS

        log = {k: v for k, v in copied.__dict__.items() if k in keep_fields}

        try:
            user_data = g.client_data
            log["requesting_client"] = user_data.client_id
            log["client_token_id"] = user_data.token_id
        except:
            pass

        try:
            log["request_id"] = request.rid
        except:
            pass

        return json.dumps(log)


__all__ = [
    "DefaultFormatter"
]



class ClientData:
    """This is a convenience class that is expected to provide basic information about a client id
    This class is meant to be extended to store any other
    client information that is desired by the developers, while still providing a consistent interface which
    lowball can use to determine client_id->roles

    """
    def __init__(self, client_id, roles, **kwargs):
        self.client_id = client_id
        self.roles = roles

    @property
    def client_id(self):
        return self._client_id

    @client_id.setter
    def client_id(self, value):
        if not isinstance(value, str):
            raise TypeError("client_id must be a non-empty string")
        if not value:
            raise ValueError("client_id must be a non-empty string")
        self._client_id = value

    @property
    def roles(self):
        return self._roles

    @roles.setter
    def roles(self, value):
        if not isinstance(value, list):
            raise TypeError("roles must be a list of strings")
        if not all(isinstance(role, str) for role in value):
            raise ValueError("roles must be a list of strings")
        self._roles = value

    def to_dict(self):
        return {
            "client_id": self._client_id,
            "roles": self.roles
        }


__all__ = [
    "ClientData"
]



from .core import Lowball
from .models import config_from_file, config_from_object
from .authentication import require_authenticated_user, require_admin, require_all_of_these_roles, \
    require_any_of_these_roles



import datetime
import json
from distutils.util import strtobool

from flask import g, current_app, request
from flask.blueprints import Blueprint

from lowball.authentication.wrappers import require_authenticated_user, require_admin
from lowball.exceptions import ListTokensInvalidReturnTypeException, ListTokensInvalidTokenInListException, \
    TokenNotFoundException, BadRequestException, InternalServerErrorException, NotFoundException, \
    MalformedProviderPackageException, InadequateRolesException, MalformedTokenIdException, \
    AuthenticationNotInitializedException, NoAuthenticationDatabaseException, NotImplementedException, \
    MalformedAuthPackageException, ListClientsInvalidReturnTypeException

from lowball.models.authentication_models import Token, ClientData, valid_token_id

auth = Blueprint("auth", __name__, url_prefix="/builtins/auth")

_JSON_HEADER = {"Content-Type": "application/json"}


@auth.route("/", methods=["POST"], strict_slashes=False)
def login():
    """Login a client to the Lowball application.

        The POST body of this request is going to be whatever is necessary to create
        the type of auth package that is configured to be used by the auth provider of
        this application. That data will be mapped into the __init__ of the auth provider
        package class. That auth package will be used to authenticate against the
        configured auth provider. The return data from that operation will be passed
        into the authenticator, which will create a token that is returned to the client.

        :rtype: tuple
        :return: token and token data, status code
        """
    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException

    if current_app.auth_db is None:
        raise NoAuthenticationDatabaseException

    post_data = request.get_json()

    auth_package_class = current_app.auth_provider.auth_package_class
    if not auth_package_class:
        raise NotImplementedException("auth_package_class",
                                      "The auth provider has not defined an authentication package")
    try:
        auth_package = auth_package_class(**post_data)
    except Exception as err:
        raise MalformedAuthPackageException(str(err))

    client_data = current_app.auth_provider.authenticate(auth_package)
    if not isinstance(client_data, ClientData):
        raise NotImplementedException("authenticate", "the auth provider is not returning a ClientData object")
    token, token_data = current_app.authenticator.create_token(client_id=client_data.client_id, roles=client_data.roles)
    current_app.logger.info(f"Successful login for {token_data.client_id}",
                            extra={"client_id": token_data.client_id, "token_id": token_data.token_id})

    current_app.auth_db.add_token(token_data)

    return json.dumps({"token": token, "token_data": token_data.to_dict()}), 200, _JSON_HEADER


@auth.route("/", methods=["DELETE"], strict_slashes=False)
@require_authenticated_user
def logout():
    """Logout a client from the Lowball application.

        This route simply revokes the token that is used to make the request.

        :rtype: tuple
        :return: empty response, status code
        """
    token_data = g.client_data
    current_app.auth_db.revoke_token(token_data.token_id)
    current_app.logger.info(f"Successfully logged out {token_data.client_id}",
                            extra={"client_id": token_data.client_id, "token_id": token_data.token_id})
    return "", 204


@auth.route("/", methods=["GET"], strict_slashes=False)
@require_authenticated_user
def whoami():
    """Get data for user token."""
    token_data = g.client_data.to_dict()
    return json.dumps(token_data), 200, _JSON_HEADER


@auth.route("/tokens", methods=["GET"], strict_slashes=False)
@require_authenticated_user
def get_current_client_tokens():
    """Get a list of all tokens in the application auth database.

        The tokens that are returned can be filtered by role, username and whether
        the token is expired. These filters are passed in the query params of the
        request. If there are multiple values passed in for username or roles, then
        any token that has any of those usernames or roles will be returned.

        The include_expired query param can be used to filter out tokens that are
        expired, but this defaults to True, so the normal operation of this view
        function is to return all tokens, both expired and not.

        :rtype: tuple
        :return: list of tokens, status code
        """

    client_data = g.client_data
    now = datetime.datetime.utcnow()
    roles = request.args.getlist("roles")
    # non admin clients cannot lookup tokens for other users

    exclude_expired = request.args.get("exclude_expired", "").lower() == "yes"

    tokens = current_app.auth_db.list_tokens_by_client_id(client_data.client_id)

    if not isinstance(tokens, list):
        raise ListTokensInvalidReturnTypeException

    if not any(isinstance(token, Token) for token in tokens):
        raise ListTokensInvalidTokenInListException

    if roles:
        tokens = [token for token in tokens if any(role in token.roles for role in roles)]

    if exclude_expired:
        tokens = [token for token in tokens if token.expiration > now]

    return json.dumps([token.to_dict() for token in tokens]), 200, _JSON_HEADER


@auth.route("/tokens", methods=["POST"], strict_slashes=False)
@require_authenticated_user
def create_token():
    """Create a new token for the specified user or self user

        Things to note:
            - A client_id is required to make a token
            - If no roles are supplied in the POST body, then the roles for the
              user will be assigned to an empty array (no roles)
            - If no token life is present in the POST body, then the configured default
              token life will be used.
            - Token life must be a positive integer greater than 9.

        :rtype: tuple
        :return: token data, status code
        """
    post_data = request.get_json()
    if post_data is None:
        post_data = {}
    default_token_life = current_app.lowball_config.authentication.default_token_life
    max_token_life = current_app.lowball_config.authentication.max_token_life
    requesting_client_data = g.client_data
    requesting_client_roles = requesting_client_data.roles
    admin_user = "admin" in requesting_client_roles
    requesting_user = requesting_client_data.client_id
    client_id = post_data.get("client_id")

    if client_id and not admin_user and requesting_client_data.client_id != client_id:
        raise InadequateRolesException("Unable to create tokens for another client")

    if not client_id:
        client_id = requesting_user

    requested_roles = post_data.get("roles", [])
    try:
        token_life = default_token_life if post_data.get("token_life") is None else int(post_data["token_life"])
    except:
        raise BadRequestException(
            f"token_life must be a positive integer greater than 10 and less than {max_token_life}")

    if not isinstance(token_life, int) or token_life < 10 or token_life > max_token_life:
        raise BadRequestException(
            f"token_life must be a positive integer greater than 10 and less than {max_token_life}")

    if not admin_user:
        if current_app.auth_provider is None:
            raise AuthenticationNotInitializedException

        try:
            target_client = current_app.auth_provider.get_client(client_id)
        except NotImplementedException as err:
            err.code = 409
            err.description = "get_client is not implemented in this auth provider. " \
                              "Non admin clients are unable to create tokens"
            raise err

        if not target_client:
            raise NotFoundException("Client ID for requesting token not found in auth provider")
        target_client_roles = target_client.roles
        if any(role not in target_client_roles for role in requested_roles):
            raise InadequateRolesException("Unable to create token with requested roles")
    roles = requested_roles

    expiration_time = datetime.datetime.utcnow() + datetime.timedelta(seconds=token_life)

    token, token_data = current_app.authenticator.create_token(client_id, roles, requesting_user, expiration_time)
    current_app.auth_db.add_token(token_data)

    return json.dumps({"token": token, "token_data": token_data.to_dict()}), 201, _JSON_HEADER


@auth.route("/tokens", methods=["DELETE"], strict_slashes=False)
@require_authenticated_user
def delete_current_client_tokens():
    """Revoke all the tokens for the specified client in the auth db."""

    client_information = g.client_data

    client_id = client_information.client_id

    tokens = current_app.auth_db.list_tokens_by_client_id(client_id)

    if not isinstance(tokens, list):
        raise ListTokensInvalidReturnTypeException

    if any(not isinstance(token, Token) for token in tokens):
        raise ListTokensInvalidTokenInListException

    for token in tokens:
        current_app.auth_db.revoke_token(token.token_id)

    return "", 204


@auth.route("/tokens/all", methods=["GET"], strict_slashes=False)
@require_admin
def get_all_tokens():
    """Get a list of all tokens in the application auth database.

        The tokens that are returned can be filtered by role, username and whether
        the token is expired. These filters are passed in the query params of the
        request. If there are multiple values passed in for username or roles, then
        any token that has any of those usernames or roles will be returned.

        The include_expired query param can be used to filter out tokens that are
        expired, but this defaults to True, so the normal operation of this view
        function is to return all tokens, both expired and not.

        :rtype: tuple
        :return: list of tokens, status code
        """

    now = datetime.datetime.utcnow()
    roles = request.args.getlist("roles")
    clients = request.args.getlist("client_ids")

    exclude_expired = request.args.get("exclude_expired", "").lower() == "yes"
    tokens = current_app.auth_db.list_tokens()
    if not isinstance(tokens, list):
        raise ListTokensInvalidReturnTypeException

    if not any(isinstance(token, Token) for token in tokens):
        raise ListTokensInvalidTokenInListException

    if roles:
        tokens = [token for token in tokens if any(role in token.roles for role in roles)]
    if clients:
        tokens = [token for token in tokens if any(client == token.client_id for client in clients)]

    if exclude_expired:
        tokens = [token for token in tokens if token.expiration > now]

    return json.dumps([token.to_dict() for token in tokens]), 200, _JSON_HEADER


@auth.route("/tokens/all", methods=["DELETE"], strict_slashes=False)
@require_admin
def delete_all_tokens():
    """Revoke all the tokens that are in the application auth database."""

    roles = request.args.getlist("roles")
    clients = request.args.getlist("client_ids")

    if not roles and not clients:
        current_app.auth_db.revoke_all()
        return "", 204

    tokens = current_app.auth_db.list_tokens()
    if not isinstance(tokens, list):
        raise ListTokensInvalidReturnTypeException

    if not any(isinstance(token, Token) for token in tokens):
        raise ListTokensInvalidTokenInListException

    if roles:
        tokens = [token for token in tokens if any(role in token.roles for role in roles)]
    if clients:
        tokens = [token for token in tokens if any(client == token.client_id for client in clients)]

    for token in tokens:
        current_app.auth_db.revoke_token(token.token_id)

    return "", 204

@auth.route("/tokens/cleanup", methods=["POST"], strict_slashes=False)
@require_admin
def cleanup_tokens():
    """Delete all expired tokens in the application auth database."""
    current_app.auth_db.cleanup_tokens()
    return "", 204


@auth.route("/tokens/<token_id>", methods=["GET"], strict_slashes=False)
@require_authenticated_user
def get_token(token_id):
    """Lookup token data for a specific token in the application auth database.

        :type token_id: str
        :param token_id: Token identifier
        :rtype: tuple
        :return: token data dictionary, status code
        """

    client_data = g.client_data

    if not valid_token_id(token_id):
        raise MalformedTokenIdException

    token = current_app.auth_db.lookup_token(token_id)

    if token is None:
        raise TokenNotFoundException
    if not isinstance(token, Token):
        raise InternalServerErrorException("auth_db.lookup_token did not return a token object")

    if "admin" not in client_data.roles and token.client_id != client_data.client_id:
        raise InadequateRolesException("Cannot lookup a token which you do not own")

    return json.dumps(token.to_dict()), 200, _JSON_HEADER


@auth.route("/tokens/<token_id>", methods=["DELETE"], strict_slashes=False)
@require_authenticated_user
def delete_token(token_id):
    """Revoke a specific token in the application auth database.

       :type token_id: str
       :param token_id: Token identifier
       :rtype: tuple
       :return: None, status code
       """

    client_data = g.client_data
    if not valid_token_id(token_id):
        raise MalformedTokenIdException

    token = current_app.auth_db.lookup_token(token_id)

    if token is None:
        raise TokenNotFoundException
    if not isinstance(token, Token):
        raise InternalServerErrorException("auth_db.lookup_token did not return a token object")

    if "admin" in client_data.roles:
        current_app.auth_db.revoke_token(token_id)
    else:
        if token.client_id != client_data.client_id:
            raise InadequateRolesException("Cannot revoke a token which you do not own")
        current_app.auth_db.revoke_token(token_id)
    return "", 204


@auth.route("/clients", methods=["GET"], strict_slashes=False)
@require_authenticated_user
def get_current_client():
    """Return client data for requesting client

    """
    client_id = g.client_data.client_id

    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException

    client_data = current_app.auth_provider.get_client(client_id)
    if not client_data or not isinstance(client_data, ClientData):
        raise NotFoundException("Client not found")

    return json.dumps(client_data.to_dict()), 200, _JSON_HEADER


@auth.route("/clients", methods=["POST"], strict_slashes=False)
@require_authenticated_user
def client_self_update():
    """Allow a client to update their client in the application.

        Allows a client to change certain attributes of their client
        using the application auth provider. The client of the
        client that is being updated is pulled from the token that
        authenticated the request. This means that you cannot update
        the client for the client with this route, otherwise
        :meth:`client_self_update` will throw an error because the
        `client` kwarg was passed in more than once.

        :rtype: tuple
        :return: updated client, status code, headers
        """
    post_data = request.get_json()
    client_id = g.client_data.client_id

    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException

    client_data = current_app.auth_provider.get_client(client_id)
    if not client_data or not isinstance(client_data, ClientData):
        raise NotFoundException("Client not found")

    client_self_update_package_class = current_app.auth_provider.self_update_client_package_class

    try:
        client_self_update_package = client_self_update_package_class(**post_data)
    except Exception as err:
        raise MalformedProviderPackageException(str(err))

    updated_client = current_app.auth_provider.client_self_update(client_self_update_package, client_id=client_id)

    if isinstance(updated_client, ClientData):
        updated_client = updated_client.to_dict()

    if isinstance(updated_client, dict):
        updated_client = json.dumps(updated_client)
    return updated_client, 200, _JSON_HEADER


@auth.route("/clients/create", methods=["POST"], strict_slashes=False)
@require_authenticated_user
def create_client():
    """Create a new user using the Lowball application auth provider."""

    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException

    post_data = request.get_json()

    create_client_package_class = current_app.auth_provider.create_client_package_class

    try:
        create_client_package = create_client_package_class(**post_data)
    except Exception as err:
        raise MalformedProviderPackageException(str(err))

    client = current_app.auth_provider.create_client(create_client_package)

    if isinstance(client, ClientData):
        client = json.dumps(client.to_dict())

    elif isinstance(client, dict):
        client = json.dumps(client)

    return client, 201, _JSON_HEADER


@auth.route("/clients/register", methods=["POST"], strict_slashes=False)
def client_registration():
    """Self-register a client in the Lowball application.

        Allow a requester with an admin token to create their own
        client with the application auth provider. The client for
        the new client will be the client that is associated to the
        token that authenticated the request. If client is passed
        in the POST body, it is ignored.

        :rtype: tuple
        :return: new client, status code, headers
        """
    post_data = request.get_json()

    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException

    client_self_registration_package_class = current_app.auth_provider.client_registration_package_class

    try:
        client_self_registration_package = client_self_registration_package_class(**post_data)
    except Exception as err:
        raise MalformedProviderPackageException(str(err))

    new_client_data = current_app.auth_provider.client_self_register(client_self_registration_package)
    if isinstance(new_client_data, ClientData):
        new_client_data = new_client_data.to_dict()

    if isinstance(new_client_data, dict):
        new_client_data = json.dumps(new_client_data)
    return new_client_data, 201, _JSON_HEADER


@auth.route("/clients/all", methods=["GET"], strict_slashes=False)
@require_admin
def get_all_clients():
    """get listing of all clients in the system. filter by roles and client_ids

    """
    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException
    roles = request.args.getlist("roles")
    clients = current_app.auth_provider.list_clients()

    if not isinstance(clients, list) or not all(isinstance(client, ClientData) for client in clients):
        raise ListClientsInvalidReturnTypeException

    if roles:
        clients = [client for client in clients if any(role in roles for role in client.roles)]

    return json.dumps([client.to_dict() for client in clients]), 200, _JSON_HEADER


@auth.route("/clients/<client_id>", methods=["GET"], strict_slashes=False)
@require_admin
def get_client(client_id):
    """Get data for the specified user.

           :type client_id: str
           :param client_id: The client_id to get data for
           :rtype: tuple
           :return: user data dictionary, status code, json content header
           """

    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException

    client = current_app.auth_provider.get_client(client_id)

    if client is None:
        raise NotFoundException(description=f"client_id '{client_id}' not found")

    if not isinstance(client, ClientData):
        raise InternalServerErrorException("auth_provider.get_client did not return a client data object")

    return json.dumps(client.to_dict()), 200, _JSON_HEADER


@auth.route("/clients/<client_id>", methods=["POST"], strict_slashes=False)
@require_admin
def update_client(client_id):
    """Update information for the specified user.

        This route simply takes whatever is in the PATCH body and
        passed it into the auth provider instance's :meth:`update_client`
        method.

        :type client_id: str
        :param client_id: The client to update.
        :rtype: tuple
        :return: updated user, status code, headers
        """

    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException

    patch_body = request.get_json()
    client = current_app.auth_provider.get_client(client_id)

    if client is None:
        raise NotFoundException(description=f"client_id '{client_id}' not found")
    update_client_package_class = current_app.auth_provider.update_client_package_class

    try:
        update_client_package = update_client_package_class(**patch_body)
    except Exception as err:
        raise MalformedProviderPackageException(str(err))

    client = current_app.auth_provider.update_client(update_client_package, client_id)
    if isinstance(client, ClientData):
        client = json.dumps(client.to_dict())

    elif isinstance(client, dict):
        client = json.dumps(client)

    return client, 200, _JSON_HEADER


@auth.route("/clients/<client_id>", methods=["DELETE"], strict_slashes=False)
@require_admin
def delete_client(client_id):
    """Delete a user from the Lowball application auth provider."""
    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException
    client = current_app.auth_provider.get_client(client_id)

    if client is None:
        raise NotFoundException(description=f"client_id '{client_id}' not found")
    current_app.auth_provider.delete_client(client_id)
    return "", 204


@auth.route("/clients/<client_id>/enable", methods=["POST"], strict_slashes=False)
@require_admin
def enable_client(client_id):
    """Enable a user using the Lowball application auth provider."""
    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException
    client = current_app.auth_provider.get_client(client_id)

    if client is None:
        raise NotFoundException(description=f"client_id '{client_id}' not found")
    current_app.auth_provider.enable_client(client_id)
    return "", 204


@auth.route("/clients/<client_id>/disable", methods=["POST"], strict_slashes=False)
@require_admin
def disable_client(client_id):
    """Disable a user using the Lowball application auth provider."""
    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException
    client = current_app.auth_provider.get_client(client_id)
    if client is None:
        raise NotFoundException(description=f"client_id '{client_id}' not found")
    current_app.auth_provider.disable_client(client_id)
    return "", 204


@auth.route("/clients/<client_id>/roles", methods=["DELETE"], strict_slashes=False)
@require_admin
def remove_client_roles(client_id):
    """remove all roles from the specified client

    """
    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException

    client = current_app.auth_provider.get_client(client_id)
    if client is None:
        raise NotFoundException(description=f"client_id '{client_id}' not found in auth provider")

    current_app.auth_provider.delete_roles(client_id, client.roles)

    return "", 204


@auth.route("/clients/<client_id>/roles/<role>", methods=["DELETE"], strict_slashes=False)
@require_admin
def remove_client_role(client_id, role):
    """remove the specified role from the specified client

    """
    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException
    client = current_app.auth_provider.get_client(client_id)
    if client is None:
        raise NotFoundException(description=f"client_id '{client_id}' not found in auth provider")

    current_app.auth_provider.delete_roles(client_id, [role])

    return "", 204


@auth.route("/clients/<client_id>/roles/<role>", methods=["POST"], strict_slashes=False)
@require_admin
def add_client_role(client_id, role):
    """add the specified role to the specified client

    """
    if current_app.auth_provider is None:
        raise AuthenticationNotInitializedException

    client = current_app.auth_provider.get_client(client_id)
    if client is None:
        raise NotFoundException(description=f"client_id '{client_id}' not found in auth provider")

    current_app.auth_provider.add_roles(client_id, [role])

    return "", 204




class TestGetStatusRoute:

    ROUTE = "/builtins/status"

    def test_auth_provider_initialized_exception_auth_db_none_produces_proper_results(self, lowball_app_no_db,
                                                                                      mock_initialized_exception):
        data = lowball_app_no_db.get(self.ROUTE)
        assert data.status_code == 200
        assert data.get_json(force=True)["name"] == lowball_app_no_db.application.lowball_config.meta.name
        assert data.get_json(force=True)["auth_db_initialized"] is False
        assert data.get_json(force=True)["auth_provider_initialized"] is False

    def test_auth_provider_initialized_non_bool_auth_db_none_produces_proper_results(self, lowball_app_no_db,
                                                                                     mock_initialized_non_bool):
        data = lowball_app_no_db.get(self.ROUTE)
        assert data.status_code == 200
        assert data.get_json(force=True)["name"] == lowball_app_no_db.application.lowball_config.meta.name
        assert data.get_json(force=True)["auth_db_initialized"] is False
        assert data.get_json(force=True)["auth_provider_initialized"] is False

    def test_auth_provider_initialized_bool_auth_db_none_produces_proper_results(self, lowball_app_no_db,
                                                                                 mock_initialized_bool):
        data = lowball_app_no_db.get(self.ROUTE)
        assert data.status_code == 200
        assert data.get_json(force=True)["name"] == lowball_app_no_db.application.lowball_config.meta.name
        assert data.get_json(force=True)["auth_db_initialized"] is False
        assert data.get_json(force=True)["auth_provider_initialized"] is mock_initialized_bool

    def test_auth_provider_initialized_exception_auth_db_not_none_produces_proper_results(self, lowball_app_no_auth_provider,
                                                                                          mock_initialized_exception):
        data = lowball_app_no_auth_provider.get(self.ROUTE)
        assert data.status_code == 200
        assert data.get_json(force=True)["name"] == lowball_app_no_auth_provider.application.lowball_config.meta.name
        assert data.get_json(force=True)["auth_db_initialized"] is True
        assert data.get_json(force=True)["auth_provider_initialized"] is False

    def test_auth_provider_initialized_non_bool_auth_db_not_none_produces_proper_results(self, lowball_app_no_auth_provider,
                                                                                         mock_initialized_non_bool):
        data = lowball_app_no_auth_provider.get(self.ROUTE)
        assert data.status_code == 200
        assert data.get_json(force=True)["name"] == lowball_app_no_auth_provider.application.lowball_config.meta.name
        assert data.get_json(force=True)["auth_db_initialized"] is True
        assert data.get_json(force=True)["auth_provider_initialized"] is False

    def test_auth_provider_initialized_bool_auth_db_not_none_produces_proper_results(self, lowball_app_mock_providers,
                                                                                     mock_initialized_bool):
        data = lowball_app_mock_providers.get(self.ROUTE)
        assert data.status_code == 200
        assert data.get_json(force=True)["name"] == lowball_app_mock_providers.application.lowball_config.meta.name
        assert data.get_json(force=True)["auth_db_initialized"] is True
        assert data.get_json(force=True)["auth_provider_initialized"] is mock_initialized_bool



import pytest

from lowball.models.authentication_models import ClientData


class TestClientData:
    # INIT TESTS
    def test_passing_non_string_for_username_at_init_raises_exception(self, invalid_username, roles):
        with pytest.raises(TypeError):
            ClientData(client_id=invalid_username, roles=roles)

    def test_passing_empty_string_for_username_at_init_raises_exception(self, roles):
        with pytest.raises(ValueError):
            ClientData(client_id="", roles=roles)

    def test_passing_non_list_for_roles_at_init_raises_exception(self, valid_username, not_list):
        with pytest.raises(TypeError):
            ClientData(client_id=valid_username, roles=not_list)

    def test_passing_invalid_list_to_roles_at_init_raises_exception(self, valid_username, invalid_roles):
        with pytest.raises(ValueError):
            ClientData(client_id=valid_username, roles=invalid_roles)

    def test_successful_init_sets_data_properly(self, valid_username, roles):
        user = ClientData(client_id=valid_username, roles=roles)
        assert user.client_id == valid_username
        assert user.roles == roles

    # SETTER TESTS
    def test_setting_username_to_non_string_after_init_raises_exception(self, user, invalid_username):
        with pytest.raises(TypeError):
            user.client_id = invalid_username

    def test_setting_username_to_empty_string_after_init_raises_exception(self, user):
        with pytest.raises(ValueError):
            user.client_id = ""

    def test_setting_roles_to_non_list_after_init_raises_exception(self, user, not_list):
        with pytest.raises(TypeError):
            user.roles = not_list

    def test_setting_roles_to_invalid_roles_after_init_raises_exception(self, user, invalid_roles):
        with pytest.raises(ValueError):
            user.roles = invalid_roles

    def test_setting_username_to_valid_username_after_init_sets_data_properly(self, user):
        user.client_id = "other"
        assert user.client_id == "other"

    def test_setting_roles_to_roles_after_init_sets_data_properly(self, user):
        user.roles = ["admin", "lead", "analyst"]
        assert user.roles == ["admin", "lead", "analyst"]

    # TO_DICT TESTS
    def test_to_dict_formats_data_properly(self, user, expected_user_to_dict_return):
        assert user.to_dict() == expected_user_to_dict_return



from pathlib import Path

from ruamel.yaml import YAML

class MetaConfig:
    """Object representation of the Meta config section of a Lowball config.

    Defines metadata about the Lowball application.

    :param name: The name of the Lowball application that is using this config defaults to "lowball"
    :type name: str
    :param base_route: The base route of the Lowball application. This route will
        be at the base of every route in the application. defaults to the root "/"
    :type base_route: str
    :param description: A description of the Lowball application that is using this
        config.
    :type description: str
    :param tags: Tags associated with this Lowball application
    :type tags: list
    """

    # TODO: Implement the ability for someone to add arbitrary metadata

    def __init__(self, name="lowball", base_route="/", description="", tags=None, **kwargs):
        self.name = name
        self.base_route = base_route
        self.description = description
        self.tags = tags

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            raise TypeError("name must be a string")
        self._name = value

    @property
    def base_route(self):
        return self._base_route

    @base_route.setter
    def base_route(self, value):
        if not isinstance(value, str):
            raise TypeError("base_route must be a string")
        if not value.startswith("/"):
            raise ValueError("base_route must start with a forward slash (/)")
        self._base_route = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        if not isinstance(value, str):
            raise TypeError("description must be a string")
        self._description = value

    @property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, value):
        if value is None:
            value = []
        if not isinstance(value, list):
            raise TypeError("tags must be a list of strings")
        if not all(isinstance(tag, str) for tag in value):
            raise ValueError("all tag values must be strings")
        self._tags = value

    def to_dict(self):
        return {
            "name": self.name,
            "base_route": self.base_route,
            "description": self.description,
            "tags": self.tags
        }


class AuthenticationConfig:
    """Object representation of the Authentication config section of a Lowball config.

    Defines the authentication settings for token generation.

    :param default_token_life: If no token life is requested when generating a token, this value
        is used
    :type default_token_life: int
    :param max_token_life: the maximum number of seconds that a token can be valid for
    :type max_token_life: int
    :param token_secret: The secret that is used to encode JWT tokens for authentication
    :type token_secret: str
    """

    def __init__(self, default_token_life=3600, max_token_life=2592000, token_secret="CHANGE_ME", **kwargs):
        self.default_token_life = default_token_life
        self.max_token_life = max_token_life
        self.token_secret = token_secret

    @property
    def default_token_life(self):
        return self._default_token_life

    @default_token_life.setter
    def default_token_life(self, value):
        if isinstance(value, str):
            try:
                value = int(value)
            except ValueError:
                pass
        if not isinstance(value, int) or isinstance(value, bool):
            raise TypeError("default_token_life must be an integer")
        self._default_token_life = value

    @property
    def max_token_life(self):
        return self._max_token_life

    @max_token_life.setter
    def max_token_life(self, value):
        if isinstance(value, str):
            try:
                value = int(value)
            except ValueError:
                pass
        if not isinstance(value, int) or isinstance(value, bool):
            raise TypeError("max_token_life must be an integer")
        self._max_token_life = value

    @property
    def token_secret(self):
        return self._token_secret

    @token_secret.setter
    def token_secret(self, value):
        if not isinstance(value, str):
            raise TypeError("token secret must be a string")
        self._token_secret = value

    def to_dict(self):
        return {
            "default_token_life": self.default_token_life,
            "max_token_life": self.max_token_life,
            "token_secret": self.token_secret,
        }


def config_from_object(config_object):
    """Generate a Config object from a dictionary object.

    :type config_object: dict
    :param config_object: dictionary representation of a config object
    :rtype: Config
    :return: Object representation of a  Lowball config
    """
    if not isinstance(config_object, dict):
        raise TypeError("config object must be a dictionary")

    meta = MetaConfig(**config_object.get("meta", {}))
    authentication = AuthenticationConfig(**config_object.get("authentication", {}))
    application = config_object.get("application")
    auth_provider = config_object.get("auth_provider")
    auth_db = config_object.get("auth_db")
    logging = config_object.get("logging")

    return Config(meta=meta, authentication=authentication, application=application, auth_provider=auth_provider,
                  auth_db=auth_db, logging=logging)


def config_from_file(filepath):
    """Generate a Config object from the data in a config file.

    Both YAML (.yaml, .yml) and JSON (.json) config files are supported by this
    method. This method will convert the data in either type of file to a dictionary
    and then pass that data to the :func:`config_from_object` method. See documentation
    for that method for information about required data.

    :type filepath: str, :class:`pathlib.Path`
    :param filepath: path to the config file on the system
    :rtype: Config
    :return: Object representation of a  Lowball config
    """
    if not isinstance(filepath, (str, Path)):
        raise TypeError("filepath must be a string or a pathlib.Path object")

    with open(filepath, "r") as config_file:
        config_data = config_file.read()

    try:
        yaml = YAML()
        config_object = dict(yaml.load(config_data))
    except:
        raise ValueError("invalid config format")

    return config_from_object(config_object)


class Config:
    """Object representation of a Lowball config.

    :param meta: Metadata about the Lowball application
    :type meta: MetaConfig
    :param authentication: Data needed for authentication operations
    :type authentication: AuthenticationConfig
    :param application: config data contained in the `application` field of the config
    :type application: dict, optional
    :param auth_provider: config data contained in the `auth_provider` field of the config
    :type auth_provider: dict, optional
    :param auth_db: config data contained in the `auth_db` field of the config
    :type auth_db: dict, optional
    :param logging: config data contained in the `logging` field of the config
    :param kwargs: Used to catch any values that are mapped into the config object that
        are not supported fields. The idea here is to make it very easy to use `Config(**config)`
        without breaking everything. These values are not used or stored.
    :type kwargs: dict
    """

    def __init__(self, meta=MetaConfig(), authentication=AuthenticationConfig(), application=None, auth_provider=None, auth_db=None, logging=None,
                 **kwargs):
        self.meta = meta
        self.authentication = authentication
        self.application = application if application is not None else {}
        self.auth_provider = auth_provider if auth_provider is not None else {}
        self.auth_db = auth_db if auth_db is not None else {}
        self.logging = logging if logging is not None else {}

    @property
    def meta(self):
        return self._meta

    @meta.setter
    def meta(self, value):
        if not isinstance(value, MetaConfig):
            raise TypeError("meta must be a MetaConfig object")
        self._meta = value

    @property
    def authentication(self):
        return self._authentication

    @authentication.setter
    def authentication(self, value):
        if not isinstance(value, AuthenticationConfig):
            raise TypeError("authentication must be a AuthenticationConfig object")
        self._authentication = value

    @property
    def application(self):
        return self._application

    @application.setter
    def application(self, value):
        if not isinstance(value, dict):
            raise TypeError("application must be a dictionary")
        self._application = value

    @property
    def auth_provider(self):
        return self._auth_provider

    @auth_provider.setter
    def auth_provider(self, value):
        if not isinstance(value, dict):
            raise TypeError("auth_provider must be a dictionary")
        self._auth_provider = value

    @property
    def auth_db(self):
        return self._auth_db

    @auth_db.setter
    def auth_db(self, value):
        if not isinstance(value, dict):
            raise TypeError("auth_db must be a dictionary")
        self._auth_db = value

    @property
    def logging(self):
        return self._logging

    @logging.setter
    def logging(self, value):
        if not isinstance(value, dict):
            raise TypeError("logging must be a dictionary")
        self._logging = value

    def to_dict(self):
        """Return a dictionary representation of the Config object."""
        return {
            "meta": self.meta.to_dict(),
            "authentication": self.authentication.to_dict(),
            "application": self.application,
            "auth_provider": self.auth_provider,
            "auth_db": self.auth_db,
            "logging": self.logging
        }


__all__ = [
    "Config",
    "MetaConfig",
    "config_from_object",
    "config_from_file",
    "AuthenticationConfig"
]



from logging import INFO
from unittest.mock import Mock

import flask
import pytest


########################
# App Context Fixtures #
########################
@pytest.fixture
def test_client():
    test_app = flask.Flask("test_app")
    test_app.config["TESTING"] = True
    test_app.logger._log = Mock()
    test_app.logger.setLevel(INFO)
    return test_app


@pytest.fixture(params=[1, 2])
def context_no_request_env(test_client, request):
    with test_client.test_request_context() as app_context:
        old_environ = flask.request.environ

        if request.param == 1:
            del flask.request.environ
        else:
            flask.request.environ = None

        yield app_context

        # Reset the environment to the old value, other wise
        # there is an error in tearing down the application.
        flask.request.environ = old_environ


@pytest.fixture(params=["USERAGENTSTRING", None])
def http_user_agent(request):
    if request.param is not None:
        return request.param


@pytest.fixture(params=["2.2.2.2", None])
def http_x_forwarded_for(request):
    if request.param is not None:
        return request.param


@pytest.fixture
def context_normal_env(test_client, http_user_agent, http_x_forwarded_for, token):
    with test_client.test_request_context() as app_context:
        if http_x_forwarded_for:
            flask.request.environ["HTTP_X_FORWARDED_FOR"] = http_x_forwarded_for

        if http_user_agent:
            flask.request.environ["HTTP_USER_AGENT"] = http_user_agent

        flask.request.environ["REMOTE_ADDR"] = "1.1.1.1"
        flask.g.client_data = token
        # flask.request.method = "GET"

        yield app_context


@pytest.fixture
def context_exception_env(context_normal_env):
    flask.g.response_is_exception = True
    flask.g.response_exception_log_data = "TESTDATANOTREALEXCEPTION"


#####################
# Response Fixtures #
#####################
@pytest.fixture
def base_response():
    return flask.Response()


@pytest.fixture
def exception_response(base_response):
    base_response.status_code = 500
    base_response.status = "500 Internal Server Error"
    return base_response



import datetime
import uuid

import jwt

from lowball.exceptions import InvalidAuthDataException, InvalidRequestingUserException, InvalidTokenLifeException, \
    InvalidTokenException, BadRequestException
from lowball.models.authentication_models import Token, valid_token_id, generate_token_id
from lowball.models.config_models import AuthenticationConfig


class Authentication:
    """Object used to generate and decode tokens used in Lowball applications.

    This object is to be used in all Lowball applications, and is in charge
    of all operations surrounding tokens. Currently, this means creation, decoding,
    and validation of tokens.

    :param config: config data needed for token operations
    :type config: AuthenticationConfig
    """

    def __init__(self, config):
        self.config = config

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, value):
        if not isinstance(value, AuthenticationConfig):
            raise TypeError("config must be a AuthenticationConfig object")
        self._config = value

    def create_token(self, client_id, roles=[], requesting_client=None, expiration=None):
        """Create a JWT token for use in Lowball applications.

        :param client_id: client id for the token
        :type client_id: str
        :param roles: list of roles to give the client
        :type roles: list(str)
        :param requesting_client: The user requesting the token
        :type requesting_client: str
        :param expiration: time of expiration for this token
        :type expiration: datetime
        :return: Token and token data
        :rtype: str, Token
        """

        if not client_id or not isinstance(client_id, str):
            raise BadRequestException("Invalid client id")

        if not roles:
            roles = []
        if not isinstance(roles, list) or not all(isinstance(r, str) for r in roles):
            raise BadRequestException("Roles must be a list of strings")

        if requesting_client is not None and not isinstance(requesting_client, str):
            raise BadRequestException("Invalid requesting client id")

        now = datetime.datetime.utcnow()

        if not requesting_client:
            requesting_client = client_id

        if expiration is None:
            expiration = now + datetime.timedelta(seconds=self.config.default_token_life)

        token_id = generate_token_id()
        token_data = Token(cid=client_id, r=roles, cts=now, ets=expiration,
                           rcid=requesting_client, tid=token_id)

        if token_data.expiration - now > datetime.timedelta(seconds=self.config.max_token_life):
            raise InvalidTokenLifeException

        token = jwt.encode(token_data.to_dict(), self.config.token_secret, algorithm="HS256")

        return token, token_data

    def decode_token(self, token):
        """Decode a Lowball token.

        :param token: JWT Lowball token string
        :type token: str
        :return: token data object
        :rtype: Token
        """
        try:
            payload = jwt.decode(token, self.config.token_secret, algorithms=["HS256"])
        except Exception as err:
            raise InvalidTokenException(str(err))
        return Token(**payload)

    def validate_token(self, token):
        """Validate a Lowball token.

        The conditions that need to be met for a Lowball token to be valid is that:
        1. The token is able to be decoded using the `decode_token` method
        2. The token is not expired

        :param token: JWT Lowball token string
        :type token: str
        :return: Whether the token is valid
        :rtype: bool
        """
        try:
            token_data = self.decode_token(token)
            now = datetime.datetime.utcnow()
            return token_data.expiration > now
        except:
            return False


__all__ = [
    "Authentication"
]



from lowball.models.provider_models.auth_provider import AuthProvider, AuthPackage
from lowball.models.authentication_models import ClientData
from lowball.exceptions import MalformedAuthPackageException, InvalidCredentialsException


class DefaultAuthProvider(AuthProvider):
    """Default Auth Provider for Lowball Applications

    This class was developed to allow for a user to make full use of a Lowball
    application without having to create their own authentication provider.

    :param username: the username that is required for authentication
    :type username: str
    :param password: the password that is required for authentication
    :type password: str
    """
    def __init__(self, username="admin", password="nimda"):
        super(DefaultAuthProvider, self).__init__()

        if not isinstance(username, str):
            raise TypeError("username must be a string")
        self._username = username

        if not isinstance(password, str):
            raise TypeError("password must be a string")
        self._password = password

    @property
    def username(self):
        """Get the username needed for authentication"""
        return self._username

    @username.setter
    def username(self, value):
        """Can't set username after init"""
        raise PermissionError("cannot set username after init")

    @property
    def password(self):
        """Get the password needed for authentication"""
        return self._password

    @password.setter
    def password(self, value):
        """Can't set password after init"""
        raise PermissionError("cannot set password after init")

    def authenticate(self, auth_package):
        """Authenticate a user.

        Using this authentication provider, a user can authenticate if the auth package
        that he passes in contains the same username and password as those configured in
        the init of this class. If they do match, then the user will be considered
        authenticated and given the admin role.

        :param auth_package: data needed to authenticate with this provider
        :type auth_package: DefaultAuthPackage
        :return: auth data
        :rtype: AuthData
        """
        if isinstance(auth_package, self.auth_package_class):
            username = auth_package.username
            password = auth_package.password
        else:
            raise MalformedAuthPackageException

        if self.username == username and self.password == password:
            return ClientData(client_id=username, roles=["admin"])
        else:
            raise InvalidCredentialsException

    def get_client(self, client_id):
        if client_id == self.username:
            return ClientData(client_id=self.username, roles=["admin"])
        return None

    @property
    def auth_package_class(self):
        """The auth package class that this class' `authenticate` method accepts."""
        return DefaultAuthPackage


class DefaultAuthPackage(AuthPackage):
    """
    Simple auth package class for use in the default auth provider.
    """
    def __init__(self, username, password, **kwargs):
        super(DefaultAuthPackage, self).__init__(**kwargs)
        self.username = username
        self.password = password

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, value):
        if not isinstance(value, str):
            raise TypeError("username must be a string")
        self._username = value

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        if not isinstance(value, str):
            raise TypeError("password must be a string")
        self._password = value


__all__ = [
    "DefaultAuthProvider",
    "DefaultAuthPackage"
]



import json

from datetime import datetime

import re
from random import choice


TOKEN_ID_PATTERN = r"[a-zA-Z0-9]{16}"
TOKEN_ID_ALPHABET = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"


def generate_token_id():
    # generates a 16 byte token consisting of a-zA-Z0-9
    return "".join(choice(TOKEN_ID_ALPHABET) for i in range(0, 16))


def valid_token_id(token_id):
    try:
        return re.fullmatch(TOKEN_ID_PATTERN, token_id) is not None
    except:
        return False


class Token:
    """Token object

    This object stores data that is needed to create the JWT tokens that are
    the primary means of authentication in Lowball microservices.

    :param cid: the client id that will be issued a token
    :type cid: str
    :param r: list of roles that the client has
    :type r: list
    :param cts: the time that the token was created
    :type cts: datetime, str
    :param ets: the time that the token expires
    :type ets: datetime, str
    :param rcid: the client id of the entity issuing the token
    :type rcid: str
    :param tid: identifier of the token that is generated
    :type tid: str
    """
    _DATE_FORMAT = "%Y-%m-%d %H:%M:%S"

    def __init__(self, cid, r, cts, ets, rcid, tid, **kwargs):
        self.client_id = cid
        self.roles = r
        self.created = cts
        self.expiration = ets
        self.issued_by = rcid
        self.token_id = tid

    def __eq__(self, other):
        return other.__class__ == self.__class__ and hash(self) == hash(other)

    def __hash__(self):
        return hash(json.dumps(self.to_dict()))

    @property
    def client_id(self):
        return self._client_id

    @client_id.setter
    def client_id(self, value):
        if not isinstance(value, str):
            raise TypeError("client_id must be a string")
        self._client_id = value

    @property
    def roles(self):
        return self._roles

    @roles.setter
    def roles(self, value):
        if not isinstance(value, list):
            raise TypeError("roles must be a list of strings")
        if not all(isinstance(role, str) for role in value):
            raise ValueError("all roles must be strings")
        self._roles = value

    @property
    def created(self):
        return self._created

    @created.setter
    def created(self, value):
        if isinstance(value, str):
            value = datetime.strptime(value, self._DATE_FORMAT)
        if not isinstance(value, datetime):
            raise TypeError("created must be a datetime object")
        self._created = value

    @property
    def expiration(self):
        return self._expiration

    @expiration.setter
    def expiration(self, value):
        if isinstance(value, str):
            value = datetime.strptime(value, self._DATE_FORMAT)
        if not isinstance(value, datetime):
            raise TypeError("expiration must be a datetime object")
        self._expiration = value

    @property
    def issued_by(self):
        return self._issued_by

    @issued_by.setter
    def issued_by(self, value):
        if not isinstance(value, str):
            raise TypeError("issued_by must be a string")
        self._issued_by = value

    @property
    def token_id(self):
        return self._token_id

    @token_id.setter
    def token_id(self, value):
        if not isinstance(value, str):
            raise TypeError("token_id must be a string")
        if not valid_token_id(value):
            raise ValueError("token_id must match the pattern [a-zA-Z0-9]{16}")
        self._token_id = value

    def to_dict(self):
        """Return a dictionary representation of a Token object."""
        return {
            "cid": self.client_id,
            "r": self.roles,
            "cts": datetime.strftime(self.created, self._DATE_FORMAT),
            "ets": datetime.strftime(self.expiration, self._DATE_FORMAT),
            "rcid": self.issued_by,
            "tid": self.token_id
        }


__all__ = [
    "Token",
    "generate_token_id",
    "valid_token_id",
    "TOKEN_ID_PATTERN"
]
